import cairo, rsvg, math
from const import *
from preferences import preferences 
import os

# A Tile is a great user interface widget because it can display things,
# be interactive in its own right and is very descriptive visually and
# textually

# ASCII art
#-----------------------------------------------------------#
# +-----+ TITLE                              Button +-----+ #
# |  A  | Description                        Button +  B  + #
# +-----+ *****                              Status +-----+ #
#-----------------------------------------------------------#
# A Icon image
# B Animated image space
# some elements could be omitted, for instance buttons or icon
# or description can easily go. 

# Todo
#  full justification

class Tile:
    def __init__( self, cr=None, render_now=False,  
                 title=None, 
                 icon=None, 
                 description=None, 
                 votes=None, 
                 rating="gold",
                 status=None,
                 buttons=[], 
                 rect=None,
                 opacity=None ):
        if not cr:
            render_now = False
        else:
            self.cr = cr
        
        self.buttons = buttons
        self.starcolour = rating
        rect = (rect.x, rect.y, rect.width, rect.height)
        self.SetRect( rect )
        self.SetMaxButtons()
        self.SetPadding()
        self.SetRating( votes )
        self.SetIcon( icon )
        
        if not render_now:
            return
        
        self.SetTitle( title )
        self.SetDescription( description )
        self.opacity = opacity
        if self.opacity == None:
            self.opacity = 100
        (status_text, status_image, frame) = status
        self.SetStatus( status_text, status_image, frame )
        
        self.DrawTile()
        
    def SetTitle( self, title=None ):
        if not title:
            return
        self.title = title
        self.title_text_ratio = 0.3
        
        self.title_size = (self.title_text_ratio * (self.height-2))
            
        self.title_y = self.title_size + self.padding /2

    def SetIcon( self, icon=None ):
        self.tile_info_x = self.height
        self.icon_height = self.height - ( 2 * self.padding )
        if not icon:
            return
        self.icon = icon
        
    def SetDescription( self, description=None ):
        if not description:
            return
        self.description_height = (self.height - (2*self.padding)) * 0.25
        self.description_width = self.width - (2* (self.height + self.padding)) - 80
        self.description = []        
        if int(self.description_height) <= 0:
            return
        
        # This should handle word wrapping correctly
        #line = None
        #description.replace("\n", "")
        #words = description.split(" ")
        #for word in words:
        #    old_line = line
        #    if line:
        #        line = line + " " + word
        #    else:
        #        line = word
        #    (x_b, y_b, text_width, t_h, x_a, y_a) = self.cr.text_extents( line )
        #    if text_width > self.description_width:
        #        self.description.append(old_line)
        #        line = word
        #self.description.append(line)
        self.description = description.split("\\n")
        line_count=len(self.description)
        self.description_ratio = line_count/self.description_height
        self.description_size = self.description_height/(line_count*.9)
        self.desc_y = float(self.title_y) + (self.description_height - float(self.description_size*(line_count-1))) + 4
        self.description_line_height = self.description_size
        if int(self.description_line_height) == 0:
            return
        self.description_rows = line_count

    def SetRating( self, rating=None ):
        self.star = preferences['prefix'] + "/pixmaps/rating."+self.starcolour+".png"
        self.rating = rating
        self.ratings_ratio = 0.3
        self.ratings_height = self.ratings_ratio * self.height
        
    def SetStatus( self, text=None, image=None, frame=None ):
        self.status_image = image
        self.status_text = text
        
        if not self.status_image:
            return
        
        status_image = preferences['prefix'] + "/pixmaps/"+self.status_image
        
        if frame != None:
            frame = "%d" % frame
            self.status_image = status_image+"."+frame+".png"
        
        if not os.path.isfile(self.status_image):
            self.status_image = status_image+".svg"
    
    def SetRect( self, rect=None ):
        if not rect:
            rect = ( 0, 0, 500, 60 )
        ( self.x, self.y, self.width, self.height ) = rect
        self.rect = rect
        
    def SetMaxButtons( self, buttons=4 ):
        self.max_buttons = buttons
        self.button_height = (self.height-(buttons+1)) / (buttons+1)
        
    def SetPadding( self, padding=5 ):
        self.padding = padding
    
    def GetRatingDimensions (self):
        tile_rating_y = self.icon_height - self.ratings_height + self.padding
        image = cairo.ImageSurface.create_from_png ( self.star )
        p_width, p_height = image.get_width(), image.get_height()
        ratings_width = (p_width/p_height)*self.ratings_height
        
        return (self.tile_info_x, tile_rating_y, 5*ratings_width, self.ratings_height)
        
    def IsOnRating (self, x, y):
        x1, y1, w, h = self.GetRatingDimensions()
        return x >= x1 and x <= x1 + w and y >= y1 and y <= y1 + h
    
    def GetRatingAtPosition(self, x, y, drag=False):
        """
        Returns the rating clicked at a specific position
        """
        
        x1, y1, w, h = self.GetRatingDimensions()
        
        rating = math.ceil((x - self.tile_info_x)/(w/5.))
        
        return rating
    
    def GetButtonAtPosition(self, x, y):
        """
        Returns the button signal at a specific position
        """
        if len(self.buttons) == 0:
            return
        self.button_x = self.width - self.height - self.padding
        by = self.button_height + (self.padding/2)
        for button in self.buttons:
            (text, signal) = button
            (x_b, y_b, text_width, t_h, x_a, y_a) = self.cr.text_extents( text )
            if (x < self.button_x - (self.padding/2)) and \
               (x > self.button_x - text_width - (self.padding/2)) and \
               (y > (by - self.button_height )) and \
               (y < by+(self.padding/2)):
                return signal
            by = by + self.button_height + 3
        return None
        
    def AddButton( self, text, signal, properties, rect ):
        button = (text, signal)
        self.buttons.append( button )
        
    def DrawTile( self ):
        self.cr.save()
        oy = self.y
        self.cr.translate( self.x, self.y )
        self.DrawIcon()
        self.DrawTitle()
        self.DrawDescription()
        self.DrawRating()
        self.DrawButtons()
        self.DrawStatus()
        
        # Underline
        self.cr.save()
        ( r, g, b ) = ( 0.2, 0.2, 0.2 )
        self.cr.set_source_rgba( r, g, b, 0.5)
        self.cr.move_to( 80, self.height - ( self.padding/2 )+0.5 )
        self.cr.line_to( self.width - 80, self.height - ( self.padding/2 )+0.5 )
        self.cr.set_line_width( 1 )
        self.cr.stroke()
        self.cr.restore()
        
        self.cr.restore()
        
        if self.opacity == 100:
            return
        
        # Fade
        self.cr.save()
        #yscroll = (self.y + 2 ) % 32
        #ypos = self.y - yscroll
        self.cr.translate( 0,self.y-2)
        self.cr.set_source_rgba( 1,1,1, 1 - (float(self.opacity)/100) )
        self.cr.rectangle(0,0, 800, 68)
        self.cr.fill()
        self.cr.restore()
        
    
    def DrawIcon( self, x=None, y=None ):
        '''
        Simple image draw function, could equally be used for any kind of image
        image can be moved by x/y
        '''
        if not self.icon:
            return
        self.cr.save()
        if x and y:
            self.cr.translate( x, y )
        
        self.cr.translate( self.padding, self.padding )
        if self.icon.endswith( ".svg" ):
            # Render SVG files
            i_svg = rsvg.Handle( self.icon )
            self.icon_width = ( float( self.icon_height )/float( i_svg.props.height ) ) * float( i_svg.props.width )
            self.cr.translate( ( self.icon_height - self.icon_width )/2, 0 )
            if i_svg.props.width > 1 and i_svg.props.height > 1 and self.icon_width > 1:
                self.cr.scale( float( self.icon_width )/float( i_svg.props.width ), 
                               float( self.icon_height )/float( i_svg.props.height ) )
                i_svg.render_cairo( self.cr )
        elif self.icon.endswith( ".png" ):
            # Render PNG files     
            image = cairo.ImageSurface.create_from_png ( self.icon )
            p_width, p_height = image.get_width(), image.get_height()
            self.icon_width = ( float( self.icon_height )/float( p_height ) ) * float( p_width )
            
            self.cr.translate( ( self.icon_height - self.icon_width )/2, 0 )
            if p_width > 1 and p_height > 1 and self.icon_width > 1:
                self.cr.scale( self.icon_width/float( p_width ), 
                               self.icon_width/float( p_height ))
            
            
            self.cr.set_source_surface( image, 0, 0 );
            self.cr.paint()
        self.cr.restore()
        
    def DrawTitle( self ):
        style = BOLD
        font = None
        effects = (None, None, CAIRO_HINT_STYLE_SLIGHT, None)
        alignment = LEFT
        color = (0,0,0)
        properties = ( self.title_size, style, font, alignment, color ) 
        self.DrawText(self.title, properties, (self.tile_info_x, self.title_y, None, None), effects)
        
    def DrawDescription( self ):
        if int(self.description_height) <= 0:
            return
        style = NORMAL
        font = None
        effects = (None, None, CAIRO_HINT_STYLE_SLIGHT, None)
        alignment = LEFT
        color = (0.3,0.3,0.3)
        properties = ( self.description_size, style, font, alignment, color ) 

        self.DrawText(self.description, properties, (self.tile_info_x, self.desc_y, None, None), effects)
        
    def DrawRating( self ):
        tile_rating_y = self.icon_height - self.ratings_height + self.padding
        
        image = cairo.ImageSurface.create_from_png ( self.star )
        p_width, p_height = image.get_width(), image.get_height()
        ratings_width = (p_width/p_height)*self.ratings_height
            
        self.cr.save()
        self.cr.translate( self.tile_info_x, tile_rating_y )
        self.cr.scale( float( ratings_width )/float( p_width ), 
                           float( self.ratings_height )/float( p_height ) )
        for x in range( 5 ):
            if x < self.rating:
                self.cr.set_source_surface( image, 0, 0 );
                self.cr.paint()
            else:
                rating_radius = 3
                self.cr.arc( p_width/2, 
                             p_height/2, 
                             rating_radius, 0, 2.0 * math.pi )
                self.cr.set_source_rgba( 0.73, 0.73, 0.65, 1)
                self.cr.fill_preserve()
                self.cr.set_source_rgba( 0.3, 0.3, 0.3, 1)
                self.cr.set_line_width( 1 )
                self.cr.stroke()
            self.cr.translate( p_width , 0 )
        self.cr.restore()
            
    def DrawButtons( self ):
        if len(self.buttons) == 0:
            return
        self.button_x = self.width - self.height
        effects = (None, None, CAIRO_HINT_STYLE_SLIGHT, None)
        properties = ( self.button_height, UNDERLINE, None, RIGHT, (.0, .0, .6))
        y = self.button_height + (self.padding/2)
        for button in self.buttons:
            (text, signal) = button
            rect = (self.button_x,y,None,None)
            self.DrawText(text, properties, rect, effects)
            y = y + self.button_height + 3
    
    def DrawStatus( self ):
        if self.status_text:
            x = self.width - self.height
            y = self.max_buttons * (self.button_height + (self.padding/2))
            rect = (x,y,None,None)
            effects = (None, None, CAIRO_HINT_STYLE_DEFAULT, None)
            properties = ( self.button_height + 3, BOLD, None, RIGHT, (.8, .8, .8))
            self.DrawText(self.status_text, properties, rect, effects)
        
        # Todo: Fix animation alignments
        if self.status_image:
            self.cr.save()
            self.cr.translate( self.width - self.height + self.padding, self.padding )

            if self.status_image.endswith( ".svg" ):
                # Render SVG files
                i_svg = rsvg.Handle( self.status_image )
                width = ( float( self.icon_height )/float( i_svg.props.height ) ) * float( i_svg.props.width )
                self.cr.translate( ( self.icon_height - self.icon_width )/2, 0 )
                if i_svg.props.width > 1 and i_svg.props.height > 1 and self.icon_width > 1:
                    self.cr.scale( float( width )/float( i_svg.props.width ), 
                                   float( self.icon_height )/float( i_svg.props.height ) )
                    i_svg.render_cairo( self.cr )
            elif self.status_image.endswith( ".png" ):
                # Render PNG files     
                image = cairo.ImageSurface.create_from_png ( self.status_image )
                p_width, p_height = image.get_width(), image.get_height()
                self.icon_width = ( float( self.icon_height )/float( p_height ) ) * float( p_width )
                
                self.cr.translate( ( self.icon_height - self.icon_width )/2, 0 )
                if p_width > 1 and p_height > 1 and self.icon_width > 1:
                    self.cr.scale( self.icon_width/float( p_width ), 
                                   self.icon_width/float( p_height ))
                self.cr.set_source_surface( image, 0, 0 );
                self.cr.paint()
            self.cr.restore()
    
    
    
    def DrawText( self, text, properties, rect=None, effects=None):
        ( size, style, font, alignment, color ) = properties 
        if effects:
            ( shadow, stroke, hinting, opacity ) = effects
            if shadow:
                ( shadow_offsetX, shadow_offsetY, shadow_opacity, shadow_size ) = shadow
            if stroke:
                ( stroke_width, stroke_color, stroke_opacity ) = stroke
            else:
                stroke_width = None
        else:
            ( shadow, stroke, hinting, opacity ) = (None, None, None, None)
            ( shadow_offsetX, shadow_offsetY, shadow_opacity, shadow_size ) = (None, None, None, None)
            ( stroke_width, stroke_color, stroke_opacity ) = (None, None, None)
        
        self.cr.save()
        if rect:
            ( x ,y, width, height ) = rect
        else:
            ( x ,y, width, height ) = ( 0, 0, None, None )
        
        self.cr.set_font_size( size )
        if not font:
            font = "Bitstream Vera Sans"
        # styles 
        if style & BOLD:
            weight = CAIRO_FONT_WEIGHT_BOLD
        else:
            weight = CAIRO_FONT_WEIGHT_NORMAL
            
        if style & ITALIC:
            slant = CAIRO_FONT_SLANT_ITALIC
        else:
            slant = CAIRO_FONT_SLANT_NORMAL
            
        self.cr.select_font_face( font, slant, weight)
        (fascent, fdescent, fheight, fxadvance, fyadvance) = self.cr.font_extents()

        try:
            lines = text.split("\n")
        except:
            lines = text

        if not y:
            y = 3 + fyadvance 
        
        # Calculate the real Y position from the supplied rectangle
        cursor_y = y - fyadvance
        
        # Work out which method we should use for hinting our fonts
        if not hinting:
            hinting =  CAIRO_HINT_STYLE_FULL

        if stroke_width and hinting == CAIRO_HINT_STYLE_STROKE:
            hinting = CAIRO_HINT_STYLE_FULL
        
        if self.opacity < 100:
            hinting = CAIRO_HINT_STYLE_MEDIUM
            
        if style & UNDERLINE:
            uline_start = x + 2
        else:
            uline_start = None
            
        for index,line in enumerate(lines):
            (x_bearing, y_bearing, text_width, text_height, x_advance, y_advance) = self.cr.text_extents( line )
            # Justification
            cursor_x = self.tile_info_x
            if alignment == RIGHT:
                cursor_x = x - text_width
                uline_end = x - text_width
                if uline_start:
                    uline_start = uline_start + 2
            elif alignment == CENTER:
                cursor_x = x + (self.tile_info_x/2)
            else:
                if uline_start:
                    uline_start = uline_start - 1
                uline_end = x + text_width
            self.cr.save()
            self.cr.move_to(cursor_x, cursor_y)
            font_opts = self.cr.get_font_options()
            font_opts.set_hint_style(CAIRO_HINT_NUM - index)
            self.cr.text_path( line )
            ( r, g, b ) = color
            self.cr.set_source_rgba( r, g, b, 1)
            #self.cr.set_source_rgba( r, g, b, 1)
            #if hinting == CAIRO_HINT_STYLE_STROKE:
            #    self.cr.fill_preserve()
            #    ( r, g, b ) = color
            #    self.cr.set_source_rgba( r, g, b, 1)
            #    self.cr.set_line_width( float(size/200) )
            #    self.cr.stroke()
            #elif hinting < CAIRO_HINT_STYLE_STROKE:
            if stroke_width:
                self.cr.fill_preserve()
                ( r, g, b ) = stroke_color
                self.cr.set_source_rgba( r, g, b, stroke_opacity)
                self.cr.set_line_width( stroke_width )
                self.cr.stroke()
            else:
                self.cr.fill()
            
            self.cr.restore()
            self.cr.save()
            if style & UNDERLINE:
                ( r, g, b ) = color
                self.cr.set_source_rgba( r, g, b, 1)
                self.cr.move_to( uline_start, cursor_y+2.5 )
                self.cr.line_to( uline_end, cursor_y+2.5 )
                self.cr.set_line_width( 0.6 )
                self.cr.stroke()
            self.cr.restore()
            cursor_y = cursor_y + fheight
        self.cr.restore()            
