import os, string, re, time

# CountInstances
import subprocess
from subprocess import PIPE

# WaitForExit
import select, thread

# WineDoors global instances
from preferences import preferences
from log import log
from const import *
from utils import GetCDMountPoint

class Wine:
    def __init__( self ):
        self.gui = True

    def CountInstances( self, process_name="wine" ):
        process_name = process_name + " "
        log.Debug( "Counting instances of " + process_name )
        ps = subprocess.Popen( ["ps", "xwww"], stdout=PIPE )
        gp = subprocess.Popen( ["grep", process_name], stdin=ps.stdout, stdout=PIPE )
        gg = subprocess.Popen( ["grep", "-v", "grep"], stdin=gp.stdout, stdout=PIPE )
        gs = subprocess.Popen( ["grep", "-v", "python"], stdin=gg.stdout, stdout=PIPE )
        wc = subprocess.Popen( ["wc", "-l"], stdin=gs.stdout, stdout=PIPE )
        processes = wc.communicate()[0]
        processes = processes.rstrip()
        log.Debug( "Found "+ processes+" " + process_name + "processes active" )
        processes = int( processes )

        return processes
    
    def CheckDrive( self, drivepath=None ):
        self.drive = False
        if not drivepath:
            drivepath = preferences['wineroot']
        log.Log( "Checking wine drive: " + drivepath )
        if not os.path.isdir( drivepath ):
            log.Debug( "wine.py: CheckDrive: Didn't find folder " + drivepath )
        else:
            log.Debug( "wine.py: CheckDrive: Found folder " + drivepath )
        
        if not os.path.isdir( os.path.expanduser( "~/.wine-doors/" ) ):
            log.Debug( "wine.py: CheckDrive: Didn't find folder ~/.wine-doors/" )
        else:
            log.Debug( "wine.py: CheckDrive: Found folder ~/.wine-doors/" )
        
        confpath = os.path.join( preferences['wineroot'], "wine-doors/configured" )
        
        if not os.path.isfile( confpath ):
            log.Debug( "wine.py: CheckDrive: Didn't find file " + confpath)
        else:
            self.drive = True
            log.Debug( "wine.py: CheckDrive: Found file " + confpath)

        if not self.drive:
            log.Log( "wine.py: CheckDrive: No wine-drive defined in specified wineroot" )
            
        return self.drive
        
        
    def CreateDrive( self ):
        if os.path.isdir( preferences['wineroot'] + "drive_c/" ):
            log.Log( "wine.py: CreateDrive: Wine drive already exists, cowardly refusing to manage already installed apps" )
            log.Debug( "         "+preferences['wineroot'] )
        else:
            log.Debug( "wine.py: CreateDrive: Wine Prefix Create "+preferences['wineroot'] )
            os.system( "wineprefixcreate --prefix %s -w > /dev/null 2>&1" % preferences['wineroot'] )
	        
        if not os.path.isdir( preferences['wineroot'] ):
            log.Error( "wine.py: CreateDrive: Failed to create wine drive" )
            return
        
        log.Debug( "wine.py: CreateDrive: Creating wine-doors paths" )
        
        if not os.path.exists( os.path.expanduser( "~/.wine-doors" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors" ) )
        if not os.path.exists( os.path.expanduser( "~/.wine-doors/installer-cache" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors/installer-cache" ) )
        if not os.path.exists( os.path.expanduser( "~/.wine-doors/packlists" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors/packlists" ) )
        if not os.path.exists( os.path.expanduser( "~/.wine-doors/apppacks" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors/apppacks" ) )
        if not os.path.exists( os.path.expanduser( "~/.wine-doors/icons" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors/icons" ) )
        log.Debug( "wine.py: CreateDrive: Configuring drives" )
        os.system( "winecfg -D" ) # autoconfigure drives
        self.WaitForExit()
        # I owe this honorable code to Mr Jason Tackaberry without whom it wouldn't have
        # been possible.
        log.Debug( "wine.py: CreateDrive: Registry changes" )
        if os.path.isfile( preferences['wineroot']+"/system.reg" ):
            lines = [ line.replace( "Change preferred owner in ~/.wine/system.reg", preferences['name'] )
                      for line in file( preferences['wineroot']+"/system.reg" ) ]
            file( preferences['wineroot']+"/system.reg", "w" ).write( "".join( lines ) )
            lines = [ line.replace( "Change preferred organization in ~/.wine/system.reg", preferences['company'] )
                      for line in file( preferences['wineroot']+"/system.reg" ) ]
            file( preferences['wineroot']+"/system.reg", "w" ).write( "".join( lines ) )
        else:
            log.Error( "wine.py: CreateDrive: System.reg does not exist" )
        
        os.system("touch "+ os.path.expanduser(os.path.join(preferences['wineroot'], "wine-doors/configured") ) )
        
        # Add any registry hacks that may be needed for normal operation
        self.Execute("regedit", preferences['prefix'] + "registry/audio-alsa.reg")
        self.WaitForExit()
        self.Execute("regedit", preferences['prefix'] + "registry/fakeie6.reg")
        self.WaitForExit()
        log.Log("wine.py: CreateDrive: Finished")
        os.system("ln -s "+self.ConvertPath(self.getProgramFilesDir())+" "+
                  preferences['wineroot']+"/drive_c/Program\ Files")
        # FIXME Move to preferences
        #os.system("chmod 000 ~/.local/share/applications/wine")
        
    def WaitForExit( self, timeout=300 ):
        reader, writer = os.pipe()
        def wait_for_wineserver(timeout):
            done = False
            log.Debug( "Waiting for wineservers to exit" )
            for x in range( 1, timeout+1 ):
                if self.CountInstances( "wineserver" ) == 0:
                    done = True
                    break
                else:
                    time.sleep( 1 )
            if not done and self.CountInstances( "wineserver" ) > 0:
                log.Error( "Wineservers still running after %d seconds" % timeout )
            else:
                os.write( writer, "done" )
        thread.start_new_thread( wait_for_wineserver, ( timeout, ) )
        readable, dummy, dummy = select.select( ( reader, ), (), (), timeout )
        if readable:
            return True
        else:
            return False
    
    def Kill( self ):
        log.Debug( "Killing wine" )
        for pname in ( "wine", "wine-pthread", "wineserver", "winedbg" ):
            while self.CountInstances( pname ):
                log.Debug( "Killing %s" % pname )
                os.system( "/usr/bin/killall -9 %s" % pname )
        self.Reboot()
    
    def Reboot( self ):
        log.Debug( "Rebooting wine" )
        os.system( "%s wineboot" % preferences['winebinary'] )
        
    def isShellScript( self, script ):
        if script.endswith( ".sh" ):
            return True
        else:
            return False
            
	    filetype = os.popen( "head -n 1 "+script ).read()
        if re.search ("^#!/", filetype):
            return True
        return False
        
    def ConvertPath( self, path ):
        """
        Convert a path from wine notation to linux absolute path
        """
        return os.popen("winepath "+ path ).readlines()[0].rstrip()
    
    def Execute( self, executable, argstring=None ):
        run = None
        if argstring == "":
            argstring = None

        if executable == "regedit" and not argstring:
            return

        if re.search("^.:\\\\", executable):
            exec_orig = executable
            executable = self.ConvertPath( executable )
            oldcwd = os.getcwd()
            os.chdir( os.path.expanduser( "~/" ) )
        else:
            oldcwd = False
        
        if not os.path.isfile(executable) and executable != "regedit":
            hit = False
            log.Warn("Can't find executable %s, Assuming CD drive path" % executable)
            cd = GetCDMountPoint()
            exec_orig = re.sub("\\\\\\\\", "/", exec_orig)
            exec_orig = re.sub("\\\\", "/", exec_orig)

            for thiscd in cd:
                log.Warn(thiscd)
                executable = re.sub("^.*:", str(thiscd), exec_orig)
                if os.path.isfile(executable):
                    hit = True
                    break

            if not hit:
                log.Warn("Executable file not found: " + executable)
                return FILE_NOT_FOUND
        
        if executable == "regedit":
            executable = "C:\\windows\\regedit.exe"
        elif re.search(".msi", executable):
            run = "msiexec /i" + executable
        elif re.search(".ahk", executable):
            run = preferences['winebinary'] + " \"" + self.getProgramFilesDir() + "\\AutoHotKey\\AutoHotKey.exe\" " + executable
        elif self.isShellScript( executable ):
            run = "/bin/sh " + executable

        if not run:
            run = "%s \"%s\"" % (preferences['winebinary'], executable)
        
        if argstring:
            run += " " + argstring

        # Set the bottle to run wine from, hide any output
        run = "WDPREFIX=" + preferences['prefix'] + " " + "WINEPREFIX=" + preferences['wineroot'] + " " + run
        
        if not preferences['debug']:
            run += " >> ~/.wine.stdout.log 2>>~/.wine.stderr.log"

        log.Debug( "Executing: " + run )
        retval = os.system( run )

        if oldcwd:
            os.chdir( oldcwd )
        return retval

    def getRegistry( self, path, key=None ):
        """
        Expects a path in the form of HKLM\\\\..\\\\.. where HKLM = System.reg
        and HKCU = User.reg, all in all this is a pretty nasty hack :/
        """
        if path.find("HKCU") != -1:
            if os.path.isfile( preferences['wineroot']+"/user.reg" ):
                reg_file = preferences['wineroot']+"/user.reg"
                path = path.replace("HKCU\\\\", "")
            else:
                log.Error("Cannot locate user.reg")
                return
        
        if path.find("HKLM") != -1:
            if os.path.isfile( preferences['wineroot']+"/system.reg" ):
                reg_file = preferences['wineroot']+"/system.reg"
                path = path.replace("HKLM\\\\", "")
            else:
                log.Error("Cannot locate system.reg")
                return
        if not key:
            reg = re.compile("(.*)\\\\(.*)\Z")
            m = reg.match(path)
            path = m.group(1)
            key = m.group(2)
        data = ""
        inkey = False
        for line in file(reg_file):
            if line.find(path) != -1:
                inkey = True
            elif inkey:
                if line.find("[") != -1:
                    break
                if line.find("=") == -1:
                    continue
                line = line.replace( "\"", "")
                line = line.replace( "\n", "")
                [thiskey, thisval] = line.split( "=")
                if thiskey == key:
                    data = thisval
                    break
        return data
          
    def getProgramFilesDir( self ):
        return self.getRegistry("HKLM\\\\Software\\\\Microsoft"+
                                "\\\\Windows\\\\CurrentVersion", 
                                "ProgramFilesDir")

wine = Wine()

if __name__ == "__main__":
    wine.CheckDrive()
    if not wine.drive:
        print "Wine Drive does not exist, creating one"
        wine.CreateDrive()

    proc = wine.CountInstances()
    if proc > 0:
       print "Counted %d wine processes, waiting for exit" % proc
       wine.WaitForExit()
